/* Copyright 2021 The Matrix.org Foundation C.I.C
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


ALTER TABLE refresh_tokens
  -- We add an expiry_ts column (in milliseconds since the Epoch) to refresh tokens.
  -- They may not be used after they have expired.
  -- If null, then the refresh token's lifetime is unlimited.
  ADD COLUMN expiry_ts BIGINT DEFAULT NULL;

ALTER TABLE refresh_tokens
  -- We also add an ultimate session expiry time (in milliseconds since the Epoch).
  -- No matter how much the access and refresh tokens are refreshed, they cannot
  -- be extended past this time.
  -- If null, then the session length is unlimited.
  ADD COLUMN ultimate_session_expiry_ts BIGINT DEFAULT NULL;
